: # use perl
eval '(exit $?0)' && eval 'exec perl -S $0 ${1+"$@"}'
  & eval 'exec perl -S $0 $argv:q'
  if 0;

#
# IBM Confidential
# OCO Source Materials
# Rational ClearQuest
# (C) Copyright IBM Corp. 2007, 2009.  All Rights Reserved.
#
# The source code for this program is not published or otherwise
# divested of its trade secrets, irrespective of what has been
# deposited with the U.S. Copyright Office.
#
##############################################################################

=head1 NAME

cqtssetup.pl

=head1 SYNOPSIS

B<cqtssetup.pl> B<[-man]> B<[-help]>

=head1 OPTIONS

=over 4

=item -help

Print a brief help message.

=item -man

Print the man page.

=back

=head1 DESCRIPTION

=head1 AUTHOR

=cut

use CQPerlExt;

use strict;
use warnings;
use IO::File;
use XML::Writer;


sub trim($);
sub getSplitedfieldList($);
sub countDbIds;
sub outputSolrCopyFieldBlock;
sub outputSolrFieldBlock;
sub outoputCQCopyRightInfo;
sub outoutDefaultSetting;
sub outputFieldLine;
sub processEntity_CQ;
sub processEntities;
sub outputFor_Entity;
sub outputFor_Solr;
sub outputFor_CQ;
sub processEntity_EntityList;
sub processEntity_Solr;


# DO NOT TOUCH!!!  Modifying any of these values could stop ClearQuest from working properly.
my $entityCreatedDateFieldName = "RATL_CQ_ENTITY_FIRST_CREATED_DATE";
my $entityLastModifiedDateFieldName = "RATL_CQ_ENTITY_LAST_MODIFIED_DATE";
my $uniqueFieldName = "RATL_CQ_UNIQUE_FIELD";
my $allFieldsName = "RATL_CQ_ALL_FIELDS";


##############################################################################
sub Usage
{
	print <<EOF;

This script is a used to create CQ Full-Text Search configuration files.

Usage: cqperl cqtssetup.pl -s <dbset>
                           -d <databasename>
                           -u <user>
                           -p <password>
                           -o <Entity | CQ | Solr>
                           [-f <input_file>]

Parameters:
  -s    A ClearQuest database set name
  -d    The ClearQuest user database name
  -u    Login name of the dbset
  -p    Password of the login name
  -o    The operation to be taken; must be "Solr", "CQ" or "Entity"
  -f    Optional 'Entity' file name.  Used only when "Solr" or "CQ" 
        operation is executing.

This script is used to generate CQ Full-Text Search configuration files.  
There are three run modes (or operations) that this command provides.  
One operation is to list all entities and their fields of a CQ database -- 
this is known as 'Entity' operation.  A second operation is to generate the 
Search Engine's configuration files (in this case Solr) -- this is known 
as 'Solr' operation.  And a third operation is to generate the CQ's 
configuration files -- this is known as 'CQ' operation.  The last two 
operations, 'Solr' and 'CQ' use the output from the 'Entity' operation to 
generate their output.

Example 1.  Generate a file listing all EntityDefs and their Fields in the 
            SAMPL database.  The generated file name will be called:
            Entity-SAMPL-TextSearch.txt
  cqperl cqtssetup.pl -s TextSarch -d SAMPL -u admin -p "" -o Entity

Example 2.  Generate the CQ configuration file to be used by CQ Full-Text 
            Search listing all EntityDefs and their Fields to be indexed and 
            searched.  The generated file name will be called:
            CQ-SAMPL-TextSearch.txt
  cqperl cqtssetup.pl -s TextSarch -d SAMPL -u admin -p "" -o CQ -f Entity-SAMPL-TextSearch.txt

Example 3.  Generate the Search Engine configuration file to be used by 
            Full-Text Search Engine (Solr in this case) listing all fields 
            to be indexed and searched.  The generated file name will be 
            called: Solr-SAMPL-TextSearch.txt
  cqperl cqtssetup.pl -s TextSarch -d SAMPL -u admin -p "" -o Solr -f Entity-SAMPL-TextSearch.txt

EOF
	exit 1;
}


if ((@ARGV == 0) || ($ARGV[0] =~ /^-[h\?]/i))
{
	Usage();
	exit 1;
}

require "CQPerlExt.pm" || die;


##############################################################################
# Process command line and setup the envirement
#

my %usage = ();


$usage{s} = "-";	# dbset
$usage{d} = "-";	# databasename
$usage{u} = "-";	# user
$usage{p} = "-";	# password
$usage{o} = "-";	# Solr | CQ | Entity
$usage{f} = "";		# file


my $validArgs = 10;

if (($#ARGV + 1) < $validArgs)
{
	print "Invalid number of argumetns ...\n\n";
	Usage();
	exit 1;
}


# Process the command line switches
for (my $switch = 0; $switch < $validArgs; $switch += 2)
{
	my ($d, $switchValue) = split("-", $ARGV[$switch]);
	
	if ($usage{$switchValue})
	{
		$usage{$switchValue} = $ARGV[$switch + 1];
	}
	else
	{
		print "Invalid option \"$ARGV[$switch]\" ...\n\n";
		Usage();
		exit 1;
	}
}


use Fcntl;
use Cwd 'abs_path';

my $path = abs_path(".");

my %entityFieldList = ();
my %replicaNamesList = ();


# Read user input arguments
for (my $inputCount = $validArgs; $inputCount < $#ARGV; $inputCount++)
{
	if ($ARGV[$inputCount] eq "-f")
	{
		# Read user input arguments from the input file
		$usage{"f"} = $ARGV[$inputCount + 1];
		
		open(inFile, $usage{"f"}) or die "Can't open input file: $!\n";
		
		my $line = "";
		while ($line = <inFile>)
		{
			my ($entity, $fields) = split("=", $line);
			
			$entity = trim($entity);
			$fields = trim($fields);
			
			if ($entity ne "")
			{
				$entityFieldList{$entity} = $fields;
			}
		}
	}
}

my $dbSet = $usage{"s"};		# -s 7.1
my $dbName = $usage{"d"};		# -d SAMPL
my $user = $usage{"u"};			# -u admin
my $pwd = $usage{"p"};			# -p ""
my $output = $usage{"o"};		# -o Solr | CQ | Entity
my $entityFile = $usage{"f"};	# <-f file-name>


print "DBSet        : $dbSet\n";
print "DBName       : $dbName\n";
print "User         : $user\n";
print "Pwd          : ********\n";
print "Output       : $output\n";
print "File         : $entityFile\n";



my $fieldFormat = "F";
my $copyFormat = "C";
my $solrOutput = "Solr";
my $cqOutput = "CQ";
my $entityOutput = "Entity";



##############################################################################
# Create session, logon, and process entities / fields
#


# Admin session, to get list of replica names
eval
{
	my $adminSessionObj = CQAdminSession::Build();
	
	$adminSessionObj->Logon($user, $pwd, $dbSet);
	my $replicName;
	my $replicaNames = $adminSessionObj->GetReplicaNames();
	foreach $replicName (@$replicaNames)
	{
		$replicaNamesList{$replicName} = "";
	}
	
	CQAdminSession::Unbuild($adminSessionObj);
};
die "$@\n" if $@; $@ = "";


# User session, for doing actual work
my $sessionObj;

eval
{
	$sessionObj = CQSession::Build();

	$sessionObj->UserLogon($user, $pwd, $dbName, $dbSet);
};
die "$@\n" if $@; $@ = "";


if (!%entityFieldList)
{
	# They didn't give us any entity / fields on the command line or file,
	# so get them from the CQ Database
	
	my $entityDefNames = $sessionObj->GetSubmitEntityDefNames();
	
	# Iterate over the record types
	foreach my $entityName ( @$entityDefNames )
	{
		my $entDef = $sessionObj->GetEntityDef($entityName);
		
		my $fieldsListString = "";
		my ($fieldNamesRef) = $entDef->GetFieldDefNames();
		foreach my $fieldName (@$fieldNamesRef)
		{
			if ($fieldName eq "history")
			{
				next;
			}
			
			$fieldsListString = $fieldsListString.",".$fieldName;
		}
		
		$entityFieldList{$entDef->GetName()} = $fieldsListString;
	}
}


# Will hold all the fields to be index by Solr
my %solrFieldList = ();

# Our XML writer object
my $xmlWriter;

# Our Entity writer object
my $entityWriter;


# Execute based on the output type request
if ($output eq $entityOutput)
{
	outputFor_Entity($output);
}
elsif ($output eq $solrOutput)
{
	outputFor_Solr($output);
}
elsif ($output eq $cqOutput)
{
	outputFor_CQ($output);
}
else
{
	print "Info         : Invalid output option: $output\n";
}


CQSession::Unbuild($sessionObj);

print "Status       : Completed\n";

exit;



##############################################################################
sub processEntities
{
	my ($outputKind) = @_;
	
	my $totalRecords = 0;
	
	# We now have a list of entities, process them
	while ( my ($entityName, $fieldList) = each (%entityFieldList) )
	{
		if ($entityName eq "history")
		{
			next;
		}
		
		eval
		{
			my $entityDefObj = $sessionObj->GetEntityDef($entityName);
			
			if ($outputKind eq $entityOutput)
			{
				$totalRecords += processEntity_EntityList($entityDefObj, $fieldList);
			}
			elsif ($outputKind eq $solrOutput)
			{
				$totalRecords += processEntity_Solr($entityDefObj, $fieldList);
			}
			elsif ($outputKind eq $cqOutput)
			{
				$totalRecords += processEntity_CQ($entityDefObj, $fieldList, $xmlWriter);
			}
		};
		warn "Warning      : Failed to process Entity '$entityName'\n" if $@; $@ = "";
	}
	
	print "Info         : Total Records ".$totalRecords."\n";
}


##############################################################################
sub outputFor_Entity
{
	my ($outputKind) = @_;

	my $theFile = "./$outputKind-$dbSet-$dbName.txt";
	print "Info         : Writing to: $theFile\n";
	open (entityWriter, ">$theFile") or die "Error      : Can't create file: $theFile\n";
	
	processEntities($outputKind);
	
	close (entityWriter);
	
	print "Info         : Wrote to: $theFile\n";
}


##############################################################################
sub outputFor_Solr
{
	my ($outputKind) = @_;
	
	my $theFile = "./$outputKind-$dbSet-$dbName.xml";
	print "Info         : Writing to: $theFile\n";
	my $xmlOutputFile = new IO::File(">$theFile") or die "Error      : Can't create file: $theFile\n";
	$xmlWriter = new XML::Writer( DATA_INDENT => "true", DATA_MODE => "true", DATA_INDENT => 2, OUTPUT => $xmlOutputFile );
	
	$xmlWriter->xmlDecl("UTF-8");
	$xmlWriter->startTag("Solr");
	
	processEntities($outputKind);
	
	$xmlWriter->comment("Copy everthing between <Solr>..</Solr> into Solr's schema.xml file");
	$xmlWriter->comment("replacing existing block with with what's in this file.");
	
	outputSolrFieldBlock();
	outputSolrUniqueField();
	outputSolrDefaultSearchField();
	outputSolrCopyFieldBlock();
	
	$xmlWriter->endTag();
	$xmlWriter->end();
	
	print "Info         : Wrote to: $theFile\n";
}


##############################################################################
sub outputFor_CQ
{
	my ($outputKind) = @_;
	
	my $theFile = "./$outputKind-$dbSet-$dbName.xml";
	print "Info         : Writing to: $theFile\n";
	my $xmlOutputFile = new IO::File(">$theFile") or die "Error      : Can't create file: $theFile\n";
	$xmlWriter = new XML::Writer( DATA_INDENT => "true", DATA_MODE => "true", DATA_INDENT => 2, OUTPUT => $xmlOutputFile );
	
	$xmlWriter->xmlDecl("UTF-16");
	
	outoputCQCopyRightInfo($xmlWriter);
	
	$xmlWriter->startTag("CQTextSearch");
		
	outoutDefaultSetting($xmlWriter);
	
	$xmlWriter->comment("List of Entitires for DBSet($dbSet), DBName($dbName)");
	$xmlWriter->comment("\'disp_order\' means to show the data of this field to the user in the order specified.");
	$xmlWriter->comment("\'index\' means should this field's data be index or not.");
	$xmlWriter->comment("The default value for those 3 attributes is \'false\', \'0\', and \'true\'");
	$xmlWriter->startTag("CQEntityList");
	
	processEntities($outputKind);
	
	$xmlWriter->endTag();
	$xmlWriter->endTag();
	$xmlWriter->end();
	
	print "Info         : Wrote to: $theFile\n";
}


##############################################################################
sub processEntity_EntityList
{
	my ($entityDefObj, $xmlWriter) = @_;
	
	my $entityDefName = $entityDefObj->GetName();
	
	my $totalDbIds = countDbIds($entityDefObj->GetDbName());
	
	print "Processing   : $entityDefName [total: ".$totalDbIds."]\n";
	
	print entityWriter "$entityDefName=";
	
	# Loop through the fields
	my ($fieldNamesRef) = $entityDefObj->GetFieldDefNames();
	foreach my $fieldName (@$fieldNamesRef)
	{
		if ($fieldName eq "history")
		{
			next;
		}
	
		print entityWriter "$fieldName,";
	}
	
	print entityWriter "\n";
	
	return $totalDbIds;
}


##############################################################################
sub processEntity_Solr
{
	my ($entityDefObj, $fields) = @_;
	
	my $entityDefName = $entityDefObj->GetName();
	
	my %fieldsList = getSplitedfieldList($fields);
	my $haveFieldsList = keys(%fieldsList);
	
	my $totalDbIds = countDbIds($entityDefObj->GetDbName());
	
	print "Processing   : $entityDefName [total: ".$totalDbIds."]\n";
	
	my $fieldNamesList = $entityDefObj->GetFieldDefNames();
	
	foreach my $fieldName (@$fieldNamesList)
	{
		if ($fieldName eq "history")
		{
			# We don't search "history" data
			next;
		}
		elsif ($fieldName eq "record_type" or $fieldName eq "version")
		{
			# Those fied types, "record_type", "dbid", and "version" are of required type
			$solrFieldList{$fieldName} = "required";
		}
		else
		{
			outputFieldLine($fieldFormat, $haveFieldsList, $fieldName, %fieldsList);
		}
	}
	
	# Force in our unique field
	$solrFieldList{$uniqueFieldName} = "required";
	
	return $totalDbIds;
}


##############################################################################
sub processEntity_CQ
{
	my ($entityDefObj, $fields, $xmlWriter) = @_;
	
	my $entityDefName = $entityDefObj->GetName();
	my $entityDefDBName = $entityDefObj->GetDbName();
	
	my %fieldsList = getSplitedfieldList($fields);
	my $haveFieldsList = keys(%fieldsList);
	
	my $totalDbIds = countDbIds($entityDefObj->GetDbName());
	
	print "Processing   : $entityDefName [total: ".$totalDbIds."]\n";
	
	my $fieldNamesList = $entityDefObj->GetFieldDefNames();
	
	$xmlWriter->startTag("CQEntity", "name" => "$entityDefName", "dbname" => "$entityDefDBName", "index" => "true");
	
	# Loop through the fields
	foreach my $fieldName (@$fieldNamesList)
	{
		if ($fieldName eq "history")
		{
			# Those are fields we don't want to index
			next;
		}
		elsif ($fieldName eq "record_type" or $fieldName eq "version" or $fieldName eq "dbid")
		{
			# Those are required fields
			$xmlWriter->startTag("field", "name" => "$fieldName", "disp_order" => "0", "index" => "true");
			$xmlWriter->endTag();
		}
		elsif (exists $fieldsList{ $fieldName })
		{
			# Those are optional user defined fields
			$xmlWriter->startTag("field", "name" => "$fieldName", "disp_order" => "0", "index" => "true");
			$xmlWriter->endTag();
		}
	}
	
	$xmlWriter->endTag();
	
	return $totalDbIds;
}


##############################################################################
sub outputFieldLine
{
	my ($format, $haveFieldsList, $fieldName, %fieldsList) = @_;
	
	if ($haveFieldsList > 0)
	{
		# We have field(s), match and add it
		if (exists $fieldsList{ $fieldName })
		{
			$solrFieldList{$fieldName} = "";
		}
	}
}


##############################################################################
sub outoutDefaultSetting
{
	my ($xmlWriter) = @_;
	
	for my $key ( keys %replicaNamesList )
	{
		my $siteName = $key;
		
		$xmlWriter->comment("Solr Search Engine default parameters");
		$xmlWriter->startTag("SearchEngine", "sitename" => $siteName, "provider" => "Solr", "version" => "1.2", "lowercasewildcard" => "true");
		
		$xmlWriter->startTag("searchrequest", "retries" => "3", "url" => "http://localhost:14080/solr/select/?q=%s&start=%d&rows=%d");
		$xmlWriter->endTag();
		$xmlWriter->startTag("searchprop", "pagesize" => "25", "cachesize" => "1000");
		$xmlWriter->endTag();
		$xmlWriter->endTag();
		
		$xmlWriter->comment("ClearQuest Database Crawler default parameters");
		$xmlWriter->startTag("CQDbCrawler", "sitename" => $siteName);
		
		$xmlWriter->comment("Number of entities to process at once and the wait (in seconds) between each cycle");
		$xmlWriter->startTag("batch", "size" => "250", "delay" => "30", "thread" => "5", "directory" => "./batch-output", "url" => "http://localhost:14080/solr/update");
		$xmlWriter->endTag();
		$xmlWriter->startTag("update", "size" => "250", "delay" => "600", "login_iteration" => "1", "directory" => "./update-output", "url" => "http://localhost:14080/solr/update");
		$xmlWriter->endTag();
		$xmlWriter->endTag();
		
		$xmlWriter->comment("ClearQuest Database and CM Server parameters");
		$xmlWriter->startTag("CQDatabase", "sitename" => $siteName);
		$xmlWriter->startTag("cmserver");
		$xmlWriter->characters("http://localhost:12080/TeamWeb/services/Team/");
		$xmlWriter->endTag();
		$xmlWriter->startTag("repository", "dbset" => $dbSet, "dbname" => $dbName, "username" => $user, "password" => "", "keyfile" => "");
		$xmlWriter->endTag();
		$xmlWriter->endTag();
	}
}


##############################################################################
sub outoputCQCopyRightInfo
{
	my ($xmlWriter) = @_;
	
	$xmlWriter->comment("");
	$xmlWriter->comment("# IBM Confidential");
	$xmlWriter->comment("# OCO Source Materials");
	$xmlWriter->comment("# Rational ClearQuest");
	$xmlWriter->comment("# (C) Copyright IBM Corp. 2007, 2009.  All Rights Reserved.");
	$xmlWriter->comment("#");
	$xmlWriter->comment("# The source code for this program is not published or otherwise");
	$xmlWriter->comment("# divested of its trade secrets, irrespective of what has been");
	$xmlWriter->comment("# deposited with the U.S. Copyright Office.");
	$xmlWriter->comment("");
	
	my $datetime = localtime();
	
	$xmlWriter->comment("Generated on $datetime");
}


##############################################################################
sub outputSolrFieldBlock
{
	$xmlWriter->startTag("fields");
	$xmlWriter->comment("Solr index fields");
	
	while ( my ($fieldName, $requirence) = each(%solrFieldList) )
	{
		if ($requirence ne "")
		{
			$xmlWriter->startTag("field", "name" => $fieldName, "type" => "string", "indexed" => "true", "stored" => "true",  "required" => "true",  "multiValued" => "false");
		}
		else
		{
			$xmlWriter->startTag("field", "name" => $fieldName, "type" => "text",   "indexed" => "true", "stored" => "false", "required" => "false", "multiValued" => "true");
		}
		
		$xmlWriter->endTag();
	}
	
	# Holds all the field data -- used for searching across all entities
	$xmlWriter->startTag("field", "name" => $allFieldsName, "type" => "text",   "indexed" => "true", "stored" => "false", "required" => "false", "multiValued" => "true");
	$xmlWriter->endTag();
	
	$xmlWriter->endTag();
}


##############################################################################
sub outputSolrUniqueField
{
	$xmlWriter->comment("The unique field to be used by Solr to identify this record / document uniquely");
	
	$xmlWriter->startTag("uniqueKey");
	$xmlWriter->characters($uniqueFieldName);
	$xmlWriter->endTag();
}


##############################################################################
sub outputSolrDefaultSearchField
{
	$xmlWriter->comment("Field for the QueryParser to use when an explicit fieldname is absent");
	
	$xmlWriter->startTag("defaultSearchField");
	$xmlWriter->characters($allFieldsName);
	$xmlWriter->endTag();
}


##############################################################################
sub outputSolrCopyFieldBlock
{
	$xmlWriter->comment("Solr copyfield block");
	
	while ( my ($fieldName, $requirence) = each(%solrFieldList) )
	{
		$xmlWriter->startTag("copyField", "source" => $fieldName, "dest" => $allFieldsName);
		$xmlWriter->endTag();
	}
}


##############################################################################
sub countDbIds
{
	my ( $entityDefName ) = @_;
	
	my $count = 0;
	
	eval
	{
		my $sqlString = "select count(*) from $entityDefName where dbid <> 0";
		
		my $resultSet = $sessionObj->BuildSQLQuery($sqlString);
		
		$resultSet->Execute();
		
		my $status = $resultSet->MoveNext();
		
		while ($status eq $CQPerlExt::CQ_SUCCESS)
		{
			$count = $resultSet->GetColumnValue(1);
			
			$status = $resultSet->MoveNext();
		}
	};
	warn "Warning      : Failed to determine min and max DBID for '$entityDefName'\n" if $@; $@ = "";
	
	return $count;
}


##############################################################################
sub getSplitedfieldList($)
{
	my $string = shift;
	
	my @fieldsList = split(",", $string);
	
	my %finalFieldsList = ();
	
	for my $field (@fieldsList)
	{
		$finalFieldsList{ trim($field) } = "";
	}
	
	return %finalFieldsList;
}


##############################################################################
sub trim($)
{
	my $string = shift;
	
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	
	return $string;
}
